#include <stdio.h>
#include <string.h>
#include <arpa/inet.h>
#include <fecpp.h>

#include <map>
#include "rsdecoder.h"
#include "rs_parm.h"

#define RS_GROUP_HISTORY 3

using namespace std;

RSDecoder::RSDecoder(Sender *sender)
	: sender(sender),
	  rs(RS_PAYLOAD_SIZE, RS_GROUP_SIZE)
{
}

void RSDecoder::send_packet(uint16_t proto, const std::string &data, uint32_t incoming_seq)
{
	int rs_group;
	if (proto == 0xffff) {
		// RS packet
		rs_group = (incoming_seq + RS_PAYLOAD_SIZE - 1) / RS_PAYLOAD_SIZE;
	} else {
		// Regular packet
		rs_group = incoming_seq / RS_PAYLOAD_SIZE;
	}

	if (rs_groups.size() >= RS_GROUP_HISTORY &&
	    rs_group < rs_groups.begin()->first) {
		// Older than the oldest group.
		return;
	}

	auto group_it = rs_groups.find(rs_group);
	if (group_it == rs_groups.end()) {
		RSGroup group;
		group.done = false;
		group_it = rs_groups.insert(make_pair(rs_group, group)).first;
	}

	RSGroup &group = group_it->second;
	if (group.done) {
		// This RS group was already sent.
		return;
	}
	if (group.packets.count(incoming_seq)) {
		// Already seen this packet.
		return;
	}

	if (proto != 0xffff) {
		sender->send_packet(proto, data, incoming_seq);
	}

	GREPacket packet;
	packet.seq = incoming_seq;
	packet.proto = proto;
	packet.data = data;
	// Don't care about ts.

	group.packets.insert(make_pair(incoming_seq, packet));
	if (group.packets.size() >= RS_PAYLOAD_SIZE) {
		// Enough to reconstruct all missing packets.

		// Reconstruction always happens on the longest packet;
		// we will truncate them later.
		int max_length = 0;
		int num_regular = 0;
		for (const auto &it : group.packets) {
			if (it.first >= rs_group * RS_PAYLOAD_SIZE) {
				// Regular packet.
				max_length = max<int>(max_length, it.second.data.size() + 4);
				++num_regular;
			} else {
				// RS packet.
				max_length = max<int>(max_length, it.second.data.size());
			}
		}

		if (num_regular < RS_PAYLOAD_SIZE) {
			// Piece the data back into the different RS groups.
			vector<string> padded_packets;
			std::map<size_t, const fecpp::byte *> shares;
			for (const auto &packet_pair : group.packets) {
				uint32_t packet_seq = packet_pair.first;
				int share_num;
				string p;
				p.resize(max_length);
				const GREPacket &packet = packet_pair.second;
				if (packet_seq >= rs_group * RS_PAYLOAD_SIZE) {
					// Regular packet.
					share_num = packet_seq - rs_group * RS_PAYLOAD_SIZE;
					uint16_t proto_be = htons(packet.proto);
					memcpy(&p[0], &proto_be, sizeof(uint16_t));
					uint16_t len_be = htons(packet.data.size());
					memcpy(&p[2], &len_be, sizeof(uint16_t));
					memcpy(&p[4], packet.data.data(), packet.data.size());
				} else {
					// RS packet.
					share_num = RS_PAYLOAD_SIZE + (rs_group * RS_PAYLOAD_SIZE - packet_seq - 1);
					memcpy(&p[0], packet.data.data(), packet.data.size());
				}
				padded_packets.push_back(p);
				shares[share_num] = reinterpret_cast<const fecpp::byte *>(padded_packets.back().data());
			}

			// Output all packets we didn't have before. They will come
			// out-of-order, which will be the job of the Reorderer to fix.
			rs.decode(shares, max_length, [&](size_t share_num, size_t num_shares, const fecpp::byte data[], size_t len){
				if (shares.count(share_num)) {
					// Already had this packet.
					return;
				}
				int packet_num = rs_group * RS_PAYLOAD_SIZE + share_num;
				uint16_t proto_be, len_be;
				memcpy(&proto_be, &data[0], sizeof(uint16_t));
				memcpy(&len_be, &data[2], sizeof(uint16_t));
				string s(&data[4], &data[4 + ntohs(len_be)]);  // TODO: security
				sender->send_packet(ntohs(proto_be), s, packet_num);
				const unsigned char *ptr = &data[4];
				printf("Reconstructed packet %d (proto=0x%04x len=%d, data=%02x %02x %02x %02x %02x %02x %02x %02x)\n", packet_num,
                                        ntohs(proto_be), ntohs(len_be),
					ptr[0], ptr[1], ptr[2], ptr[3], ptr[4], ptr[5], ptr[6], ptr[7]);
				++num_recovered_packets;
			});
		}
		
		group.done = true;
	}

	if (rs_groups.size() > RS_GROUP_HISTORY) {
		const auto &it = rs_groups.begin();
		if (!it->second.done) {
			printf("Giving up reconstruction within group [%d,%d> (only got %d/%d packets, needed %d)\n",
			       it->first * RS_PAYLOAD_SIZE,
			       (it->first + 1) * RS_PAYLOAD_SIZE,
			       int(it->second.packets.size()),
			       RS_GROUP_SIZE,
			       RS_PAYLOAD_SIZE);
		}
		rs_groups.erase(it);
	}
}

